/* linklist.c - doubly linked list manipulation

  Copyright (C) 1998, 2002, 2003 Martin A. Olevitch
  Written by Martin A. Olevitch (marty@cosray.wustl.edu)

  This file is part of Linklist.

  Linklist is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Linklist is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Linklist, see the file COPYING; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>	/* malloc */
#include <string.h>	/* memset */

#define FREE	0
#define INUSE	1

/* BADID - list ID is invalid */
#define BADID(i) ((i) >= Max || (i) < 0 || Status[(i)] != INUSE)

/* EMPTY - empty list (having no data elements) */
#define EMPTY(i)	(Count[(i)] == 0)

/* ATFIRST - first element in the list (right after the head */
#define ATFIRST(i)	(Listp[(i)]->cur == Listp[(i)]->head->next)

/* ATLAST - last element in the list (right before the tail */
#define ATLAST(i)	(Listp[(i)]->cur == Listp[(i)]->tail->prev)

/* An empty list looks like this (cur points to NULL):
 * 
 *           head      tail
 *           |         |
 *           |         |
 *           v         v
 * 
 * NULL <--- prev <--- prev
 *           NULL      NULL
 *           next ---> next ---> NULL
 * 
 * A list with a few elements looks like this (note that cur could be 
 * pointing to any element, except head or tail):
 * 
 *           head                cur                 tail
 *           |                   |                   |
 *           |                   |                   |
 *           v                   v                   v
 * 
 * NULL <--- prev <--- prev <--- prev <--- prev <--- prev
 *           NULL      data      data      data      NULL
 *           next <--- next ---> next ---> next ---> next ---> NULL
 */

struct node {
    void *p;	/* user data */
    struct node *prev;
    struct node *next;
};

struct list {
    struct node *head;
    struct node *tail;
    struct node *cur;
};

#define NLISTS	10	/* number of lists to add when we need more */

static struct list **Listp = NULL;
static int *Status = NULL;
static int *Count = NULL;
static int Max = 0;

static int alloc_list(void);

/* ll_init - initialize more lists. Returns 0 if all went well, else some
 * negative value. This can be used if the programmer knows he will need a
 * large number of lists, instead of letting the library just grow the
 * number of lists in smaller chunks. */
int
ll_init(max)
    int max;	/* the number of additional lists to initialize */
{
    int i;
    int n;

    if (max < 0) {
	return -1;
    }

    n = (Max + max) * sizeof(struct list);
    Listp = realloc(Listp, n);
    if (Listp == NULL) {
	return -2;
    }

    n = (Max + max) * sizeof(int);
    Status = realloc(Status, n);
    if (Status == NULL) {
	return -3;
    }
    for (i=Max; i < (Max+max); i++) {
	Status[i] = FREE;
    }

    Count = realloc(Count, n);
    if (Count == NULL) {
	return -4;
    }
    for (i=Max; i < (Max+max); i++) {
	Count[i] = 0;
    }

    Max += max;
    return 0;
}

/* ll_new_list - create a new linked list. Returns a non-negative list ID
 * if all went well, else some negative number. */
int
ll_new_list(void)
{
    int i;
    struct list *lp;
    int ret;

    for (i=0; i<Max; i++) {
	if (Status[i] == FREE) {
	    break;
	}
    }
    if (i >= Max) {
	/* need to add more lists - ll_init also changes Max, etc. */
	ret = ll_init(NLISTS);
	if (ret) {
	    return -1;
	}
	/* now i should be the correct index of a FREE list */
	if (Status[i] != FREE) {
	    return -5;	/* bizarre */
	}
    }

    lp = malloc(sizeof(struct list));
    if (lp == NULL) {
	return -2;
    }

    lp->head = malloc(sizeof(struct node));
    if (lp->head == NULL) {
	return -3;
    }
    lp->tail = malloc(sizeof(struct node));
    if (lp->tail == NULL) {
	return -4;
    }
    lp->cur = NULL;

    lp->head->prev = NULL;
    lp->head->p = NULL;
    lp->head->next = lp->tail;

    lp->tail->prev = lp->head;
    lp->tail->p = NULL;
    lp->tail->next = NULL;

    Listp[i] = lp;
    Status[i] = INUSE;
    Count[i] = 0;
    return(i);
}

/* ll_first - move to the first element with data of the list. If the list
 * is empty, we point to the tail of the list. Returns 0 if all went well, 
 * else returns -1. */
int
ll_first(id)
    int id;
{
    if (BADID(id)) {
	return -1;
    }
    if (EMPTY(id)) {
	return -2;
    }
    Listp[id]->cur = Listp[id]->head->next;
    return 0;
}

/* ll_last - move to the last element of the list. If the list is empty,
 * moves to the head of the list. Returns 0 if all went well, else
 * returns -1. */
int
ll_last(id)
    int id;
{
    if (BADID(id)) {
	return -1;
    }
    if (EMPTY(id)) {
	return -2;
    }
    Listp[id]->cur = Listp[id]->tail->prev;
    return 0;
}

/* ll_prev - move to previous link in the list. Returns 0 if we are already
 * at the head of the list, 1 if we did the move, -1 if the ID is bad, -2
 * if there are no data elements in the list. */
int
ll_prev(id)
    int id;
{
    if (BADID(id)) {
	return -1;
    }
    if (EMPTY(id)) {
	return -2;
    }
    if (ATFIRST(id)) {
	return 0;
    }
    Listp[id]->cur = Listp[id]->cur->prev;
    return 1;
}

/* ll_next - move to next link in the list. Returns 0 if we are already
 * at the tail of the list, 1 if we did the move, -1 if the ID is bad, or
 * -2 if there are no data elements in the list. */
int
ll_next(id)
    int id;
{
    if (BADID(id)) {
	return -1;
    }
    if (EMPTY(id)) {
	return -2;
    }
    if (ATLAST(id)) {
	return 0;
    }
    Listp[id]->cur = Listp[id]->cur->next;
    return 1;
}

/* ll_val - return a pointer to the data stored in the current node.
 * Returns NULL if the ID is bad. Note that the data pointer could itself
 * be a NULL pointer. */
void *
ll_val(id)
    int id;
{
    if (BADID(id)) {
	return NULL;
    }
    if (EMPTY(id)) {
	return NULL;
    }
    return (Listp[id]->cur->p);
}

/* ll_insert_before - add a node to the list before the current node.
 * Returns 0 if all went well, else a negative value. The new node then
 * becomes the current node. */
int
ll_insert_before(id, p)
    int id;
    void *p;	/* pointer to data */
{
    struct node *tmp;

    if (BADID(id)) {
	return -1;
    }

    tmp = malloc(sizeof(struct node));
    if (tmp == NULL) {
	return -2;
    }

    if (EMPTY(id)) {
	Listp[id]->cur = Listp[id]->tail;
    }

    tmp->next = Listp[id]->cur;
    tmp->prev = Listp[id]->cur->prev;
    tmp->p = p;

    Listp[id]->cur->prev->next = tmp;
    Listp[id]->cur->prev = tmp;
    Listp[id]->cur = tmp;

    Count[id]++;

    return 0;
}

/* ll_insert_after - add a node to the list after the current node.
 * Returns 0 if all went well, else a negative value. The new node then
 * becomes the current node. */
int
ll_insert_after(id, p)
    int id;
    void *p;	/* pointer to data */
{
    struct node *tmp;

    if (BADID(id)) {
	return -1;
    }

    tmp = malloc(sizeof(struct node));
    if (tmp == NULL) {
	return -2;
    }

    if (EMPTY(id)) {
	Listp[id]->cur = Listp[id]->head;
    }

    tmp->next = Listp[id]->cur->next;
    tmp->prev = Listp[id]->cur;
    tmp->p = p;

    Listp[id]->cur->next->prev = tmp;
    Listp[id]->cur->next = tmp;
    Listp[id]->cur = tmp;

    Count[id]++;

    return 0;
}

/* ll_insert - backward compatibility */
int
ll_insert(id, p)
    int id;
    void *p;	/* pointer to data */
{
    return ll_insert_after(id, p);
}

/* ll_delete - delete the current node. The node before it becomes the
 * current node. Returns a pointer to the data pointer so that it can be
 * free'd by the caller. */
void *
ll_delete(id)
    int id;
{
    struct node *np;
    void *p;
    int atfirst = 0;

    if (BADID(id)) {
	return NULL;
    }
    if (EMPTY(id)) {
	return NULL;
    }
    if (ATFIRST(id)) {
	/* need to check this here, because later cur will be gone */
	atfirst = 1;
    }

    np = Listp[id]->cur;
    p = np->p;

    np->next->prev = np->prev;
    np->prev->next = np->next;
    Count[id]--;

    if (EMPTY(id)) {
	Listp[id]->cur = NULL;
    } else if (atfirst) {
	Listp[id]->cur = np->next;
    } else {
	Listp[id]->cur = np->prev;
    }

    free(np);
    return p;
}

/* ll_destroy_list - free all memory in the list. Returns a negative value
 * if there were problems. Otherwise, returns a non-negative integer which
 * gives the number of data nodes freed. The pointers to the data are
 * returned via the parameter p so they can be freed by the calling
 * routine. */
int
ll_destroy_list(id, p)
    int id;
    void **p;	/* pointer to array of data pointers */
{
    int n = 0;

    if (BADID(id)) {
	return -1;
    }

    while (!EMPTY(id)) {
	*p = ll_delete(id);
	p++;
	n++;
    }
    free(Listp[id]->head);
    free(Listp[id]->tail);
    free(Listp[id]);
    Status[id] = FREE;
    return n;
}

int 
ll_val_next(int id, void **p)
{
    if (BADID(id)) {
	return -1;
    }
    if (EMPTY(id)) {
	return -1;
    }
    *p = Listp[id]->cur->p;

    return (ll_next(id));
}
